#include <assert.h>
#include <stdio.h>
# ifdef __STDC__
#include <stdlib.h>
# endif
#include <string.h>
#include <ctype.h>
# define efopen fopen
# define efclose fclose
# define eprintf printf
#include "form.h"
/* #define DEBUG_OUTPUT 1 */
/* #define DEBUG 1 */
/* #define DEBUG_EVAL 1 */

# ifdef __STDC__
extern int random(void);
extern int srandom(unsigned int);
#else
extern char *malloc();
# endif

char *cmnd;

#define MIN(x,y)	((x) < (y) ? (x) : (y))
#define MAX(x,y)	((x) > (y) ? (x) : (y))
#define ABS(x)		((x) > 0 ? (x) : (-(x)))

#define NR_FORMS	500
#define	NR_GROUPS	32

/* Try to get a uniformely distributed random number from [0..x) */
# ifdef notdef
#define RANDOM_MAX 0x80000000
#if RANDOM_MAX*NR_FORMS/NR_FORMS == RANDOM_MAX
#	define xrandom(x)  ((int)((long)(x) * random() / ((long)RANDOM_MAX + 1)))
#else 
#	define xrandom(x)  ((x) <= 1 ? 0 : (random()/(RANDOM_MAX/((x)-1))))
#endif
# endif /* notdef */

#define xrandom(x) (random() % (x))


/* How many people do w want per group? */
#define DRAIN_GRPSTUDENTS	14
#define MIN_GRPSTUDENTS		29
#define MAX_GRPSTUDENTS		31

/* Penalties for wrong number of students:
 * These implement an S-like funktion:
 * empty groups are very good, groups less than half full get better if they
 * are drained. Groups more than half full get better if they are 
 * filled, and crowded groups again get better if they lose members
 */

#define FORM_LIMIT 3	/* Max. Nr. of forms to exchange in 1 step */
#define MAX_GROUPS_PER_MOVE 3 /* 2 for the normal exchange policy */
/* Higher probability for forms in wrong groups */
#define WISH_PROB 1
#define NONWISH_PROB 2
#define REJECT_PROB 7
/* if defined this gives 1/probability for moving all members of a group */
#define TRANSFER_ALL	  100
#define CROWD_PENALTY    -900
#define NOTFULL_PENALTY   900
#define EMPTY_REWARD     1000
#define FULL_REWARD      1000
#define MAX_TRIES	 5000 /* Nr of tries until increasing water */
/* EXCHANGE / (EXCHANGE + REFILL) is probability for exchange */
#define EXCHANGE	1
#define REFILL		10

/* If the following is defined, we get a plateau for non-full
 * non-empty groups.
 */
/* #define LOW_PLATEAU 5000 */

/* Penalties for people in wrong group */
#define REJECT_PENALTY   700

/* Rewards for people in wish group */
#define WISH_REWARD	 100
#define WISH_REWARDSTEP	   0 
#define NONWISH_PENALTY    0

int wish_rewardstep;
int	nr_forms, nr_groups=1; /* There is always a 0 group */
form_t	forms [NR_FORMS];

form_t	groups [NR_GROUPS]; /* use full struct here - easier to code */
int	group_students [NR_GROUPS];
int	group_nr_forms [NR_GROUPS];
long	group_eval [NR_GROUPS];
long	group_probab[NR_GROUPS];
long	total_eval;
# ifdef THRESHOLD
long    threshold = 0;
# else /* !THRESHOLD */
long	watermark;
long	increment=1; /* how fast the water rises */
# endif /* !THRESHOLD */

#define FIRST_OF_GROUP(index) groups[index].next
#define INSERT_AFTER(group, form) do{ \
			    (form)->prev = (group);\
			    if((group)->next)\
				(group)->next->prev = (form);\
			    (form)->next = (group)->next;\
			    (group)->next = (form);\
			}while(0)
#define REMOVE(element)	do{ \
			    assert((element)->prev);\
			    if((element)->next)\
				(element)->next->prev = (element)->prev;\
			    (element)->prev->next = (element)->next;\
			    (element)->prev=(element)->next=NULL;\
			}while(0)
#define MOVE_FROM_TO_N(from,to,n)	do{ \
			    int i;\
			    form_t *last, *f=(from), *t=(to);\
			    if(!n)\
				break;\
			    assert(f);\
			    assert(t);\
			    assert(f->prev);\
			    for(last=f, i=0; i<(n)-1; i++)\
			    {\
				last=last->next;\
				assert(last);\
			    }\
			    if(last->next)\
				last->next->prev = f->prev;\
			    f->prev->next = last->next;\
			    f->prev = t;\
			    if(t->next)\
				t->next->prev = last;\
			    last->next = t->next;\
			    t->next = f;\
			}while(0)

char *strsave(s)
    char *s;
{
    char *s2;
    if((s2 = malloc(strlen(s) + 1)) == NULL)
    {
	fprintf(stderr, "Cannot malloc\n");
	exit(17);
    }
    return(strcpy(s2, s));
}

void read_forms (formfile)
    char *formfile;
{
	FILE	*fp;
	char	s[256];
	char	*p;

	fp = efopen (formfile, "r");
	if(fp == NULL)
	{
	    fprintf(stderr, "Cannot open %s\n",formfile);
	    exit(23);
	}

	while (fgets (s, sizeof (s), fp)) {
		if (strncmp (s, "ZETTEL ", 7) == 0) {
			int	i;

			nr_forms ++;
			assert (nr_forms < NR_FORMS);
			p = s + 7;
			for (i = 0; i < NR_WISH; i++) {
				forms[nr_forms-1].wish[i] = strtol(p, &p, 10);
				if(forms[nr_forms-1].wish[i] >= nr_groups)
				    nr_groups = forms[nr_forms-1].wish[i]+1;
				assert(nr_groups < NR_GROUPS);
			}
			for (i = 0; i < NR_REJECT; i++) {
				forms[nr_forms-1].reject[i] = strtol(p, &p, 10);
				if(forms[nr_forms-1].reject[i] >= nr_groups)
				    nr_groups = forms[nr_forms-1].reject[i]+1;
				assert(nr_groups < NR_GROUPS);
			}
		} else if (strncmp (s, "STUDENT ", 8) == 0) {
			/* Remove white space at end of student data */
			int len = strlen(s), i;
			for(i=len-1; i>=0; i--)
			{
			    if(!isspace(s[i]))
				break;
			    s[i] = '\0';
			}
			forms[nr_forms-1].studinfo[forms[nr_forms-1].
			    nr_students] = strsave(s+8);
			forms[nr_forms-1].nr_students ++;
		} else if (strcmp (s, "\n") != 0) {
			exit (1);
		}
	}
	efclose (fp);
}


/* flag: update form->cur_group */
int eval (group, flag)
    int group, flag;
{
	long	val = 0;
	form_t	*form;
	int	i;

#define DRAIN_BASE    ((DRAIN_GRPSTUDENTS - MIN_GRPSTUDENTS) * \
			NOTFULL_PENALTY + FULL_REWARD)

# ifdef DEBUG_EVAL
	printf("eval group %d\n",group);
# endif /* DEBUG_EVAL */
	if(group_students [group] <= MAX_GRPSTUDENTS &&
	    group_students [group] >= MIN_GRPSTUDENTS)
	{
	    val = FULL_REWARD;
	}
# ifdef LOW_PLATEAU
	else if(group_students[group] == 0) {
	    val = EMPTY_REWARD;
	} else if (group_students [group] > MAX_GRPSTUDENTS) {
	    val = (group_students [group] - MAX_GRPSTUDENTS) * CROWD_PENALTY 
		+ FULL_REWARD;
	} else 
	    val = LOW_PLATEAU;
# else /* !LOW_PLATEAU */
	else if (group_students [group] > MAX_GRPSTUDENTS) {
	    val = (group_students [group] - MAX_GRPSTUDENTS) * CROWD_PENALTY 
		+ FULL_REWARD;
	} else if (group_students [group] < DRAIN_GRPSTUDENTS) {
	    val = group_students [group] * (DRAIN_BASE - EMPTY_REWARD) /
	    DRAIN_GRPSTUDENTS + EMPTY_REWARD;
	} else if (group_students [group] < MIN_GRPSTUDENTS) {
	    val = (group_students [group] - MIN_GRPSTUDENTS) *
		NOTFULL_PENALTY + FULL_REWARD;
	}
# endif /* ! LOW_PLATEAU */

# ifdef DEBUG_EVAL
	printf("group_eval: %d stud:%d\n",val,group_students[group]);
# endif /* DEBUG_EVAL */
	group_probab[group] = 0;
	for (form = FIRST_OF_GROUP(group); form; form = form ->next) {

	    form->cur_probab = 1; /* 1<= x <= NR_WISH */
	    /* Treat group 0 as reject group for everybody */
	    if(!group)
	    {
		val -= NONWISH_PENALTY;
		goto nextone;
	    }

	    for (i = 0; i < NR_WISH; i ++) {
		if (group == form ->wish [i]) {
		    val += (WISH_REWARD - i * wish_rewardstep) *
			form->nr_students;
		    form->cur_probab += i * WISH_PROB; /* 1<= x <= NR_WISH */
		    goto nextone;
		}
	    }

	    /* It's not in a wishgroup */
	    val -= NONWISH_PENALTY * form->nr_students;
	    form->cur_probab += NR_WISH * WISH_PROB + NONWISH_PROB;

	    for (i = 0; i < NR_REJECT; i ++) {
		if (group == form ->reject [i]) {
		    val -= REJECT_PENALTY * form->nr_students;
		    form->cur_probab += REJECT_PROB;
		    goto nextone;
		}
	    }

	nextone:;
	    assert(form->cur_probab);
	    group_probab[group] += form->cur_probab;
	    if(flag)
		form->cur_group = group;
	    assert(group_probab[group]);
	}
	return val;
}


void init_groups () {
	int	i;
# ifdef notdef
	int	group;
# endif

	for(i=0; i<NR_GROUPS; i++)
	{
	    groups[i].next = groups[i].prev = NULL;
	}
# ifdef notdef
	/* Initially put all forms in their preferred group */
	for (i = 0; i < nr_forms; i ++) {

	    forms [i] .cur_group = group = forms [i] .wish [0];
	    if(!group) /* If no wish put him in random group */
	    {
		do{
		    forms [i] .cur_group = group =  xrandom(nr_groups);
		}while(!group); /* Avoid group 0 */
	    }

	    assert ((unsigned)group < nr_groups);

	    INSERT_AFTER(groups + group, forms + i);

	    group_students [group] += forms [i] .nr_students;
	    group_nr_forms[group]++;
	}
# endif /* notdef */
# ifdef notdef
	/* Put them in random groups */
	for (i = 0; i < nr_forms; i ++) {
	    do{
		forms [i] .cur_group = group =  xrandom(nr_groups);
	    }while(!group); /* Avoid group 0 */
	    assert ((unsigned)group < nr_groups);

	    INSERT_AFTER(groups + group, forms + i);

	    group_students [group] += forms [i] .nr_students;
	    group_nr_forms[group]++;
	}
# endif /* notdef */
/* # ifdef notdef */
	/* Put all in group 0 */
	for (i = 0; i < nr_forms; i ++) {
	    forms [i] .cur_group = 0;

	    INSERT_AFTER(groups, forms + i);

	    group_students [0] += forms [i] .nr_students;
	    group_nr_forms[0]++;
	}
/* # endif *//* notdef */

	for (i = 0; i < nr_groups; i ++) {
		group_eval [i] = eval (i, 1);
		total_eval += group_eval [i];
	}

# ifndef THRESHOLD
	if (total_eval > 0) {
		/* 9/10 of total_eval */
		watermark = total_eval * 9 / 10;
	} else {
		/* 1/10 below total_eval */
		watermark = total_eval + total_eval / 10;
	}
# endif /* !THRESHOLD */
}


void print_group (i, wishscore, rejscore)
    int i, *wishscore, *rejscore;
{
	int	j, found_wish; 
	form_t	*form, *lastform = NULL;
	int 	s = 0;
	static FILE *studlist = NULL;

	if(!studlist && (studlist = fopen("studliste","w")) == NULL)
	{
	    fprintf(stderr,"Cannot open studliste file\n");
	    exit(5);
	}
	printf ("%d: %d students, ==> %ld\n", i, group_students [i], group_eval [i]);
	for (form = FIRST_OF_GROUP(i); form; form = form ->next) {

		assert (!lastform || form ->prev == lastform);

		found_wish = 0;
		printf ("\t+ ");
		for (j = 0; j < NR_WISH; j ++) {
			printf ("%2d ", form ->wish [j]);
			if(!found_wish && form->wish[j] == i)
			{
			    wishscore[j] += form->nr_students;
			    found_wish++;
			}
		}
		printf ("- ");
		for (j = 0; j < NR_REJECT; j ++) {
			printf ("%2d ", form ->reject [j]);
			if(!found_wish && form->reject[j] == i)
			{
			    *rejscore += form->nr_students;
			    found_wish++;
			}
		}
		printf (": %d\n", form ->nr_students);
		for (j = 0; j < form->nr_students; j ++) {
		    assert(form->studinfo[j]);
		    fprintf(studlist,"%s %d\n", form->studinfo[j], i);
		}

		s+= form ->nr_students;
		lastform = form;
	}
	assert (s == group_students [i]);
}


void print_groups () {
	int	i; 
	int wishscore[NR_WISH];
	int rejscore = 0;
	int fullscore = 0;

	for (i = 0; i < NR_WISH; i ++) {
	    wishscore[i] = 0;
	}
	for (i = 0; i < nr_groups; i ++) {
	    print_group (i, wishscore, &rejscore);
	}
	for (i = 0; i < NR_WISH; i ++) {
	    printf("%d. Wunsch: %2d\n", i+1, wishscore[i]);
	    fullscore += wishscore[i];
	}
	printf("Ablehnung: %2d\n", rejscore);
	if (feof (stdout) || feof (stderr)) {
	    eprintf ("%s: cannot write to stdout\n", cmnd);
	    exit (1);
	}
	printf ("total: %ld Zuordnung: %d\n", total_eval, fullscore);
}

int		grpind[NR_GROUPS], grps[NR_GROUPS], grp_n[NR_GROUPS];
int		grp_students[NR_GROUPS], grp_neweval[NR_GROUPS];
int		grp_oldeval[NR_GROUPS];

void move_them(int gcount)
{
    int i,n;
    form_t *p, *next;
    /* Move first all first, second, .. wishes then nonwishes until
     * group full or nothing more in group 0 
     */
    for(n=gcount-1; n>=0; n--)
    {
	int ready=0;

#       ifdef DEBUG
	    printf("n=%d\n",n);
#       endif /* DEBUG */
	/* This group is empty afterwards */
	if(!xrandom(TRANSFER_ALL))
	    ready=1;
	for(i=0; i<NR_WISH && !ready; i++)
	{
#	    ifdef DEBUG
		printf("moving\n");
#	    endif /* DEBUG */
	    for(p=FIRST_OF_GROUP(0), next=p?p->next:NULL; p && !ready;
		p=next, next=next?next->next:NULL)
	    {
		if(p->wish[i] == grpind[n] && p->nr_students +
		    group_students[grpind[n]] <= MAX_GRPSTUDENTS)
		{
#		    ifdef DEBUG
			printf("%X  group:%d\n",p,p->cur_group);
			fflush(stdout);
#		    endif /* DEBUG */
		    group_nr_forms[0] --;
		    group_students[0] -= p->nr_students;
		    group_nr_forms[grpind[n]] ++;
		    group_students[grpind[n]] += p->nr_students;
		    assert(group_nr_forms[0] >= 0);
		    assert(group_students[0] >= 0);
		    REMOVE(p);
		    INSERT_AFTER(&groups[grpind[n]], p) ;
		}
		if(group_students[grpind[n]] >= MAX_GRPSTUDENTS ||
		    (group_students[grpind[n]] >= MIN_GRPSTUDENTS &&
		    xrandom(2)))
		{
		    ready=1;
		}
	    }
	    /* break with a fifty fifty chance after each wish */
	    if(xrandom(2))
		ready = 1;
#	    ifdef DEBUG
		printf("\n");
#	    endif /* DEBUG */
	}
	for(p=FIRST_OF_GROUP(0), next=p?p->next:NULL; p && !ready;
	    p=next, next=next?next->next:NULL)
	{
	    int good;

	    good = 1;
	    for(i=0; i<NR_REJECT; i++)
	    {
		if(p->reject[i] == grpind[n])
		{
		    good = 0;
		    break;
		}
	    }
	    if(!good)
		continue;

	    if(p->nr_students + group_students[grpind[n]] <= MAX_GRPSTUDENTS)
	    {
		group_nr_forms[0] --;
		group_students[0] -= p->nr_students;
		group_nr_forms[grpind[n]] ++;
		group_students[grpind[n]] += p->nr_students;
		assert(group_nr_forms[0] >= 0);
		assert(group_students[0] >= 0);
		REMOVE(p);
		INSERT_AFTER(&groups[grpind[n]], p) ;
	    }
	    if(group_students[grpind[n]] >= MAX_GRPSTUDENTS ||
		(group_students[grpind[n]] >= MIN_GRPSTUDENTS &&
		xrandom(2)))
	    {
		ready=1;
	    }
	}
    }
}

# ifdef THRESHOLD
# define SLID_MAX 40 /* Nr of items in sliding window */
# else /* !THRESHOLD */
# define SLID_MAX  10 /* Nr of items in sliding window */
# endif /* !THRESHOLD */

void onemove_exchange (oldeval, neweval)
    long	*oldeval, *neweval;
{
    int		ngroups, gcount=0, transfer_all=0;
    int		i, j, n, index;
    form_t	*formptr_from, *formptr_to, tmpform;

    for(i=0; i<NR_GROUPS; i++)
	grpind[i] = grps[i] = grp_n[i] = grp_students[i] = 0;
# ifdef DEBUG
    printf("Group eval vorher:\n");
    for(i=0; i<nr_groups; i++)
    {
	printf("%8d  ", group_eval[i]);
	assert(eval(i, 0) == group_eval[i]);
    }
    printf("\nin rand:%d\n",(MIN(nr_groups, MAX_GROUPS_PER_MOVE)-1));
# endif /* DEBUG */
    /* less probable for higher nr. */
    /* ngroups = xrandom(MIN(nr_groups, MAX_GROUPS_PER_MOVE)-1) + 2; */
    ngroups = xrandom(MIN(nr_groups, MAX_GROUPS_PER_MOVE)-1) *
	      xrandom(MIN(nr_groups, MAX_GROUPS_PER_MOVE)-1) /
	      MAX((MIN(nr_groups, MAX_GROUPS_PER_MOVE) - 2), 1) + 2;
    *oldeval = total_eval;

# ifdef TRANSFER_ALL
    if((transfer_all = (xrandom(TRANSFER_ALL) == 0)))
    {
	ngroups = 2;
    }
# endif /* TRANSFER_ALL */

# ifdef DEBUG
    printf("groups:%d: ",ngroups);
    fflush(stdout);
# endif /* DEBUG */
    for(gcount=0; gcount<ngroups; gcount++)
    {
	do{ /* Min 1 in at least 1 group */
	    /* no longer avoid group 0 */
	    grpind[gcount] = xrandom(nr_groups);
	}while(grps[grpind[gcount]] || /* Do not select same group twice */
	    (!gcount && !FIRST_OF_GROUP(grpind[gcount])));
	grps[grpind[gcount]] = 1;
# ifdef DEBUG
	printf("%d ",grpind[gcount]);
	fflush(stdout);
# endif /* DEBUG */
	if(transfer_all)
	{
	    if(gcount==0)
		grp_n[gcount] = group_nr_forms[grpind[gcount]];
	    else
		grp_n[gcount] = 0;
	}
	else
	{
	    /* Nr. of forms to transfer (at least 1, at most all) */
	    if(!gcount)
	    {
		grp_n[gcount] = xrandom(MIN(group_nr_forms[grpind[gcount]],
		    FORM_LIMIT))+1;
	    }
	    else
	    {
		grp_n[gcount] = xrandom(MIN(group_nr_forms[grpind[gcount]],
		    FORM_LIMIT)+1);
	    }
	}
# ifdef DEBUG
	printf("t:%d ",grp_n[gcount]);
	fflush(stdout);
# endif /* DEBUG */
    }
# ifdef DEBUG
    printf("\n Vorher:\n");
    for(n=0; n<gcount; n++)
    {
	form_t *p, *lp;
	printf("%d:",n);
	for(p=lp=groups + grpind[n]; p; ) 
	{
	    printf("%X ",p);
	    p=p->next;
	    assert(!p || p->prev == lp);
	    lp=p;
	}
	printf("\n");
    }
    printf("\n");
    fflush(stdout);
    for(n=0; n<gcount; n++)
    {
	printf("stud:%d n:%d grpstud:%d\n",group_students[grpind[n]],
	    group_nr_forms[grpind[n]], grp_students[n]);
    }
# endif /* DEBUG */

    tmpform.prev = tmpform.next = NULL;

    for(n=0; n<gcount; n++)
    {
	for(i=0; i<grp_n[n]; i++)
	{
	    /* which one to transfer next */
	    index = xrandom(group_probab[grpind[n]]);
	    formptr_to = n ? groups + grpind[n-1] : &tmpform;
	    for(formptr_from = FIRST_OF_GROUP(grpind[n]),
		j=formptr_from->cur_probab;
		j<=index; j+=formptr_from->cur_probab)
	    {
		formptr_from = formptr_from->next;
		assert(formptr_from);
	    }

	    REMOVE(formptr_from);
	    INSERT_AFTER (formptr_to, formptr_from);
	    group_students [grpind[n]] -= formptr_from->nr_students;
	    group_nr_forms[grpind[n]]--;
	    group_probab[grpind[n]] -= formptr_from->cur_probab;
	    grp_students[n] += formptr_from->nr_students;
	    if(n)
	    {
		group_students[grpind[n-1]] += formptr_from->nr_students;
		group_nr_forms[grpind[n-1]]++;
	    }
	}
    }
    MOVE_FROM_TO_N(tmpform.next, groups + grpind[gcount-1], grp_n[0]);
    group_nr_forms[grpind[gcount-1]] += grp_n[0];
    group_students[grpind[gcount-1]] += grp_students[0];
# ifdef DEBUG
    printf("Nachher:\n");
    for(n=0; n<gcount; n++)
    {
	form_t *p,*lp;
	printf("%d:",n);
	for(p=lp=groups + grpind[n]; p;) 
	{
	    printf("%X ",p);
	    p=p->next;
	    assert(!p || p->prev == lp);
	    lp=p;
	}
	printf("\n");
    }
    printf("\n");
    for(n=0; n<gcount; n++)
    {
	printf("stud:%d n:%d grpstud:%d\n",group_students[grpind[n]],
	    group_nr_forms[grpind[n]], grp_students[n]);
    }
    fflush(stdout);
# endif /* DEBUG */

    *neweval = total_eval;
# ifdef DEBUG
    printf("eval vorher %d\n", *neweval);
    printf("Group eval vorher:\n");
    for(i=0; i<nr_groups; i++)
	printf("%8d  ", group_eval[i]);
    printf("\n");
    fflush(stdout);
# endif /* DEBUG */
    for(n=0; n<gcount; n++)
    {
	grp_oldeval[n] = group_eval[grpind[n]];
	grp_neweval[n] = eval(grpind[n], 0);
	*neweval -= grp_oldeval[n];
	*neweval += grp_neweval[n];
    }
# ifdef DEBUG
    printf("eval nachher %d\n", *neweval);
    fflush(stdout);
# endif /* DEBUG */

# ifdef THRESHOLD
    /* We let the exchange happen if it does'nt degrade too much */
    if (*neweval - *oldeval >= -threshold)
# else /* !THRESHOLD */
    /* let it happen if not deeper under water */
    if (*neweval >= watermark || *neweval >= *oldeval)
# endif /* !THRESHOLD */
    {
# ifdef DEBUG
	printf("its ok\n");
	fflush(stdout);
# endif /* DEBUG */
	total_eval = *neweval;
	for(n=0; n<gcount; n++)
	{
	    group_eval[grpind[n]] = grp_neweval[n];
	    /* This assertion has side-effect of updating the groups !! */
	    assert(grp_neweval[n] == eval(grpind[n], 1));
	}
    } else {
# ifdef DEBUG
	printf("Zug zuruecknehmen\n");
	fflush(stdout);
# endif /* DEBUG */
	/* getting deeper below the water or over threshold -- undo it */
	assert(!tmpform.next);
	for(n=gcount-1; n>=0; n--)
	{
	    if(n == gcount-1)
	    {
		MOVE_FROM_TO_N((groups + grpind[n])->next, &tmpform, grp_n[0]);
	    }
	    else
	    {
		MOVE_FROM_TO_N((groups + grpind[n])->next,
		    groups + grpind[n+1], grp_n[n+1]);
	    }
	    group_students[grpind[n]] -= grp_students[(n+1)%gcount];
	    group_students[grpind[(n+1)%gcount]] += grp_students[(n+1)%gcount];
	    group_nr_forms[grpind[n]] -= grp_n[(n+1)%gcount];
	    group_nr_forms[grpind[(n+1)%gcount]] += grp_n[(n+1)%gcount];
	}
	MOVE_FROM_TO_N(tmpform.next,(groups + grpind[0]), grp_n[0]);
	assert(!tmpform.next);
# ifdef DEBUG
	printf("\n Nach zuruecknehmen:\n");
	for(n=0; n<gcount; n++)
	{
	    form_t *p,*lp;
	    printf("%d:",n);
	    for(p=lp=groups + grpind[n]; p;) 
	    {
		printf("%X ",p);
		p=p->next;
		assert(!p || p->prev == lp);
		lp=p;
	    }
	    printf("\n");
	}
	printf("\n");
	fflush(stdout);

	printf("Group eval nach rueckgaengig:\n");
	for(i=0; i<nr_groups; i++)
	    printf("%8d  ", eval(i, 0));
	printf("\nvorher:\n");
	for(i=0; i<nr_groups; i++)
	    printf("%8d  ", group_eval[i]);
	printf("\n");
	fflush(stdout);
# endif /* DEBUG */
	/* Reevaluate because eval has side-effect of updating cur_probab */
	for(n=0; n<gcount; n++)
	{
	    assert(grp_oldeval[n] == eval(grpind[n], 1));
	}
    }
}

void onemove_refill (oldeval, neweval)
    long	*oldeval, *neweval;
{
    int		ngroups, gcount=0;
    int		i, n, old_0_eval, new_0_eval, old_0_n, old_0_students;
    form_t	*p, *next;


    for(i=0; i<NR_GROUPS; i++)
	grpind[i] = grps[i] = grp_n[i] = grp_students[i] = 0;
# ifdef DEBUG
    printf("Group eval vorher:\n");
    for(i=0; i<nr_groups; i++)
    {
	printf("%8d  ", group_eval[i]);
	assert(eval(i, 0) == group_eval[i]);
    }
    printf("\nin rand:%d\n",(MIN(nr_groups, MAX_GROUPS_PER_MOVE)-1));
# endif /* DEBUG */
    /* less probable for higher nr. */
    /* ngroups = xrandom(MIN(nr_groups, MAX_GROUPS_PER_MOVE)-1) + 2; */
    ngroups = xrandom(MIN(nr_groups, MAX_GROUPS_PER_MOVE)-1) *
	      xrandom(MIN(nr_groups, MAX_GROUPS_PER_MOVE)-1) /
	      MAX((MIN(nr_groups, MAX_GROUPS_PER_MOVE) - 2), 1) + 2;
    *oldeval = *neweval = total_eval;
    old_0_n = group_nr_forms[0];
    old_0_students = group_students[0];

# ifdef DEBUG
    printf("groups:%d\n",ngroups);
    fflush(stdout);
# endif /* DEBUG */
    for(gcount=0; gcount<ngroups; gcount++)
    {
	do{
	    /* avoid group 0 */
	    grpind[gcount] = xrandom(nr_groups - 1) + 1;
	}while(grps[grpind[gcount]]);
	grps[grpind[gcount]] = 1;
# ifdef DEBUG
	{
	    form_t *p, *lp;
	    printf("gcount:%d group:%d\n",gcount,grpind[gcount]);
	    for(p=lp=groups + grpind[gcount]; p; ) 
	    {
		printf("%X  ",p);
		p=p->next;
		assert(!p || p->prev == lp);
		lp=p;
	    }
	    printf("\n");
	}
	printf("\n");
	fflush(stdout);
	{form_t *p; int count=0;
	    for(p=(groups[grpind[gcount]].next); p; p=p->next)
	    {
		count ++;
	    }
	    printf("%d %d\n",count, group_nr_forms[grpind[gcount]]);
	    fflush(stdout);
	    assert(count == group_nr_forms[grpind[gcount]]);
	}
# endif /* DEBUG */
	grp_n[gcount] = group_nr_forms[grpind[gcount]];
	assert(grp_n[gcount] >= 0);
	grp_students[gcount] = group_students[grpind[gcount]];
# ifdef DEBUG
	printf("grp:%d stud:%d n:%d\n",
	    grpind[gcount],grp_students[gcount],grp_n[gcount]);
	fflush(stdout);
# endif /* DEBUG */
	/* Move all in 0 group */
	MOVE_FROM_TO_N(FIRST_OF_GROUP(grpind[gcount]), groups,
	    group_nr_forms[grpind[gcount]]);
	group_nr_forms[0] += group_nr_forms[grpind[gcount]];
	group_students[0] += group_students[grpind[gcount]];
	group_nr_forms[grpind[gcount]] = group_students[grpind[gcount]] = 0;
    }
# ifdef DEBUG
    for(n=0; n<gcount; n++)
    {
	    form_t *p, *lp;
	    printf("gcount:%d group:%d\n",n,grpind[n]);
	    for(p=lp=groups + grpind[n]; p; ) 
	    {
		printf("%X  ",p);
		p=p->next;
		assert(!p || p->prev == lp);
		lp=p;
	    }
	    printf("\n");
	    fflush(stdout);
	}
# endif /* DEBUG */
    
    /* Do it twice there is some random stuff in it... */
    move_them(gcount);
    move_them(gcount);

# ifdef DEBUG
    printf("Nachher:\n");
    for(n=0; n<gcount; n++)
    {
	form_t *p,*lp;
	printf("%d group:%d\n",n,grpind[n]);
	for(p=lp=groups + grpind[n]; p;) 
	{
	    printf("%X  ",p);
	    p=p->next;
	    assert(!p || p->prev == lp);
	    lp=p;
	}
	printf("\n");
    }
    printf("\n");
    for(n=0; n<gcount; n++)
    {
	printf("grp:%d stud:%d n:%d grpstud:%d\n",grpind[n],
	    group_students[grpind[n]], group_nr_forms[grpind[n]],
	    grp_students[n]);
    }
    fflush(stdout);
# endif /* DEBUG */

# ifdef DEBUG
    printf("eval vorher %d\n", *neweval);
    fflush(stdout);
# endif /* DEBUG */
    for(n=0; n<gcount; n++)
    {
	grp_oldeval[n] = group_eval[grpind[n]];
	grp_neweval[n] = eval(grpind[n], 0);
	*neweval -= grp_oldeval[n];
	*neweval += grp_neweval[n];
    }
    old_0_eval = group_eval[0];
    new_0_eval = eval(0, 0);
    *neweval -= old_0_eval;
    *neweval += new_0_eval;
# ifdef DEBUG
    printf("eval nachher %d\n", *neweval);
    fflush(stdout);
# endif /* DEBUG */

# ifdef THRESHOLD
    /* We let the exchange happen if it does'nt degrade too much */
    if (*neweval - *oldeval >= -threshold)
# else /* !THRESHOLD */
    /* let it happen if not deeper under water */
    if (*neweval >= watermark || *neweval >= *oldeval)
# endif /* !THRESHOLD */
    {
# ifdef DEBUG
	printf("its ok\n");
	fflush(stdout);
# endif /* DEBUG */
	total_eval = *neweval;
	for(n=0; n<gcount; n++)
	{
	    group_eval[grpind[n]] = eval(grpind[n], 1);
	}
	group_eval[0] = eval(0, 1);
    } else {
# ifdef DEBUG
	printf("Zug zuruecknehmen\n");
	fflush(stdout);
# endif /* DEBUG */
	/* getting deeper below the water or over threshold -- undo it */
	for(n=0; n<gcount; n++)
	{
	    MOVE_FROM_TO_N(FIRST_OF_GROUP(grpind[n]), groups,
		group_nr_forms[grpind[n]]);
	}
	for(n=0; n<gcount; n++)
	{
	    for(p=FIRST_OF_GROUP(0), next=p?p->next:NULL; p;
		p=next, next=next?next->next:NULL)
	    {
		if(p->cur_group == grpind[n])
		{
		    REMOVE(p);
		    INSERT_AFTER(&groups[grpind[n]], p) ;
		}
	    }
	    group_students[grpind[n]] = grp_students[n];
	    group_nr_forms[grpind[n]] = grp_n[n];
	}
	group_students[0] = old_0_students;
	group_nr_forms[0] = old_0_n;
# ifdef DEBUG
	printf("\n Nach zuruecknehmen:\n");
	for(n=0; n<gcount; n++)
	{
	    form_t *p,*lp;
	    printf("%d group:%d\n",n,grpind[n]);
	    for(p=lp=groups + grpind[n]; p;) 
	    {
		printf("%X  ",p);
		p=p->next;
		assert(!p || p->prev == lp);
		lp=p;
	    }
	    printf("\n");
	}
	printf("\n");
	fflush(stdout);

	printf("Group eval nach rueckgaengig:\n");
	for(i=0; i<nr_groups; i++)
	    printf("%8d  ", eval(i, 0));
	printf("\nvorher:\n");
	for(i=0; i<nr_groups; i++)
	    printf("%8d  ", group_eval[i]);
	printf("\n");
	printf("grp:%d stud:%d n:%d\n",
	    grpind[n], group_students[grpind[n]],group_nr_forms[grpind[n]]);
	fflush(stdout);
# endif /* DEBUG */
	/* Reevaluate because eval has side-effect of updating cur_probab */
	for(n=0; n<gcount; n++)
	{
	    assert(grp_oldeval[n] == eval(grpind[n], 1));
	}
	eval(0, 1);
    }
}

int onemove () {
    long oldeval, neweval = -300000;
    static long	sliding_win[SLID_MAX], sliding_sum = 0;
    static int	win_index = 0, winfull = 0;
    int exchange = EXCHANGE, refill = REFILL, weight = abs(neweval / 7000);
# ifdef THRESHOLD
    static int tries = 0;
# endif /* THRESHOLD */
    
    weight = weight > 0 ? weight : 1;
    exchange = neweval < 0 ? 1 : weight;
    refill   = neweval < 0 ? weight : 1;
    if(xrandom(exchange + refill) <= exchange - 1)
	onemove_exchange(&oldeval, &neweval);
    else
	onemove_refill(&oldeval, &neweval);

    /* This implements a sliding window sum over
     * the eval for the threshold scheme
     * the differences between eval and watermark for the other
     */
# ifdef THRESHOLD
    if ((neweval - oldeval) > -threshold || tries >= MAX_TRIES) {
# endif /* THRESHOLD */
	if(winfull)
	    sliding_sum -= sliding_win[win_index];
# ifdef THRESHOLD
	sliding_sum += (sliding_win[win_index++] = (tries >= MAX_TRIES ?
	    0 : (neweval-oldeval)));
#	ifdef DEBUG_OUTPUT
	    printf("sum:%9ld tries:%5d ", sliding_sum, tries);
	    printf ("total = %8ld, threshold = %8ld\n", total_eval, threshold);
	    fflush(stdout);
#	endif /* DEBUG_OUTPUT */
	tries = 0;
# else /* !THRESHOLD */
	sliding_sum += (sliding_win[win_index++] = neweval - watermark);
# endif /* !THRESHOLD */
	if(win_index >= SLID_MAX - 1)
	{
	    winfull = 1;
	    win_index = 0;
	}
# ifdef THRESHOLD
    }
    else
    {
	tries++;
    }
    /* Lower threshold if things aren't getting better */
    if(winfull && sliding_sum <= 0)
    {
	tries = 0;
	winfull = 0;
	win_index = 0;
	sliding_sum = 0;
	if(threshold == 0)
	    return(0);
	threshold = threshold * 9 / 10;
    }
    return(1);
# else /* !THRESHOLD */
	/* Set the new increment according to difference between watermark
	 * and value of eval
	 */
	if(winfull)
	    increment = MAX(1, sliding_sum / 1000);
    { int t=(neweval >= watermark); return(t); }
# endif /* !THRESHOLD */
}

int main (argc,argv)
    int argc;
    char **argv;
{
	extern char *optarg;
	extern int optind;
	extern int getopt(int, char * const [], const char *);
	int errflg = 0, c;
	char *eptr;

	if((cmnd = strrchr(argv[0], '/')) != NULL)
	    cmnd++;
	else
	    cmnd = argv[0];
	wish_rewardstep = WISH_REWARDSTEP;
	while((c = getopt(argc, argv, "w:")) != EOF)
	{
	    switch(c)
	    {
		case 'w':
		    wish_rewardstep = strtol(optarg, &eptr, 10);
		    if(*eptr != '\0' || wish_rewardstep < 0 ||
			wish_rewardstep > 10)
		    {
			fprintf(stderr, "%s: required: 0 <= rewardstep <= 10\n",
			    cmnd);
			errflg++;
		    }
		    break;
		case '?':
		    errflg++;
		default:
		    assert(0);
	    }
	}
	if(errflg || argc - optind != 2)
	{
	    fprintf(stderr, "%s: Usage: %s [-w rewardstep] <randseed> <file>\n",
		cmnd, cmnd);
	    exit(23);
	}
	srandom(atoi(argv[optind]));
	read_forms (argv[optind + 1]);
	init_groups ();
	/* print_groups (); */
# ifdef THRESHOLD
	for (threshold = 7000; ; ) 
	    if(!onemove())
		break;
# else /* !THRESHOLD */
	/* increase watermark only if last move over watermark */
	for (;watermark < total_eval + 20 * increment; watermark += increment) {
	    int j;
	    for(j=0; j<MAX_TRIES; j++)
	    {
		if(onemove ())
		{
		    break;
		}
	    }
# ifdef DEBUG_OUTPUT
	    /* if(watermark > 50000) */
	    {
		printf("%5d tries -- ",j);
		printf ("total = %8ld, watermark = %8ld increment = %3ld\n",
		    total_eval, watermark, increment);
		fflush(stdout);
	    }
# endif /* DEBUG_OUTPUT */
	}
# endif /* !THRESHOLD */
	print_groups ();

	return 0;
}
